package com.tzld.longarticle.recommend.server.service.recall;

import com.google.common.collect.Lists;
import com.tzld.longarticle.recommend.server.common.ThreadPoolFactory;
import com.tzld.longarticle.recommend.server.mapper.crawler.CrawlerBaseMapper;
import com.tzld.longarticle.recommend.server.model.dto.Content;
import com.tzld.longarticle.recommend.server.model.dto.ContentHisPublishArticle;
import com.tzld.longarticle.recommend.server.model.entity.aigc.CrawlerMetaArticle;
import com.tzld.longarticle.recommend.server.model.entity.crawler.AccountAvgInfo;
import com.tzld.longarticle.recommend.server.model.entity.crawler.AccountCorrelation;
import com.tzld.longarticle.recommend.server.model.entity.crawler.Article;
import com.tzld.longarticle.recommend.server.model.entity.crawler.ArticleDetailInfo;
import com.tzld.longarticle.recommend.server.remote.AIGCRemoteService;
import com.tzld.longarticle.recommend.server.repository.aigc.CrawlerMetaArticleRepository;
import com.tzld.longarticle.recommend.server.repository.crawler.AccountAvgInfoRepository;
import com.tzld.longarticle.recommend.server.repository.crawler.AccountCorrelationRepository;
import com.tzld.longarticle.recommend.server.repository.crawler.ArticleDetailInfoRepository;
import com.tzld.longarticle.recommend.server.repository.crawler.ArticleRepository;
import com.tzld.longarticle.recommend.server.service.AccountIndexAvgViewCountService;
import com.tzld.longarticle.recommend.server.service.recall.strategy.DefaultRecallStrategy;
import com.tzld.longarticle.recommend.server.util.CommonCollectionUtils;
import com.tzld.longarticle.recommend.server.util.DateUtils;
import com.tzld.longarticle.recommend.server.util.JSONUtils;
import com.tzld.longarticle.recommend.server.util.Md5Util;
import com.tzld.longarticle.recommend.server.util.feishu.FeishuMessageSender;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecallService implements ApplicationContextAware {

    @Autowired
    ArticleRepository articleRepository;
    @Autowired
    AccountAvgInfoRepository accountAvgInfoRepository;
    @Autowired
    ArticleDetailInfoRepository articleDetailInfoRepository;
    @Autowired
    CrawlerMetaArticleRepository crawlerMetaArticleRepository;
    @Autowired
    AIGCRemoteService aigcRemoteService;
    @Autowired
    AccountIndexAvgViewCountService accountIndexAvgViewCountService;
    @Autowired
    CrawlerBaseMapper crawlerBaseMapper;
    @Autowired
    AccountCorrelationRepository accountCorrelationRepository;

    private final Map<String, RecallStrategy> strategyMap = new HashMap<>();
    private ApplicationContext applicationContext;
    private final ExecutorService pool = ThreadPoolFactory.recallPool();

    @Value("${recall.content.his.feishu.enable:false}")
    private Boolean contentHisFeishuEnable;
    @Value("${morning.noon.fission.rate:0.64}")
    private double morningNoonFissionRate;


    @PostConstruct
    public void init() {
        Map<String, RecallStrategy> type = applicationContext.getBeansOfType(RecallStrategy.class);
        for (Map.Entry<String, RecallStrategy> entry : type.entrySet()) {
            RecallStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }
    }

    public RecallResult recall(RecallParam param) {
        List<RecallResult.RecallData> results = new ArrayList<>();
        log.info("RecallParam {}", JSONUtils.toJson(param));
        List<RecallStrategy> strategies = getRecallStrategy(param);
        log.info("RecallStrategy {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies,
                s -> s.getClass().getSimpleName())));
        List<Content> content = getAllContent(param);
        if (CollectionUtils.isEmpty(content)) {
            return new RecallResult(results);
        }
        param.setContent(content);
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<RecallResult.RecallData>> recallResultFutures = new ArrayList<>();
        for (final RecallStrategy strategy : strategies) {
            Future<RecallResult.RecallData> future = pool.submit(() -> {
                try {
                    return strategy.recall(param);
                } finally {
                    cdl.countDown();
                }
            });
            recallResultFutures.add(future);
        }
        try {
            cdl.await(10000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("recall error", e);
            return null;
        }

        for (Future<RecallResult.RecallData> f : recallResultFutures) {
            try {
                RecallResult.RecallData data = f.get();
                results.add(data);
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }

        return new RecallResult(results);
    }

    private List<RecallStrategy> getRecallStrategy(RecallParam param) {
        List<RecallStrategy> strategies = new ArrayList<>();
        strategies.add(strategyMap.get(DefaultRecallStrategy.class.getSimpleName()));
        return strategies;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }

    private List<Content> getAllContent(RecallParam param) {
        List<Content> content = aigcRemoteService.getAllContent(param);
        if (CollectionUtils.isEmpty(content)) {
            FeishuMessageSender.sendWebHookMessage("07026a9f-43f5-448b-ba40-a8d71bd6e634", "内容召回失败\n"
                    + "ghId: " + param.getGhId() + "\n"
                    + "账号名称: " + param.getAccountName());
            return content;
        }
        // 标题历史均值
        setTitleAvgViewCount(content, param.getGhId());
        // category 查询
        setContentCategory(content);
        return content;
    }

    public void setContentCategory(List<Content> contentList) {
        long start = System.currentTimeMillis();
        Map<String, String> articleMd5Map = new HashMap<>();
        List<String> md5List = new ArrayList<>();
        for (Content content : contentList) {
            String md5 = generateArticleUniqueMd5(content.getCrawlerLink());
            md5List.add(md5);
            articleMd5Map.put(content.getId(), md5);
        }
        List<CrawlerMetaArticle> categoryList = getByUniqueIndexIn(md5List);
        if (CollectionUtils.isEmpty(categoryList)) {
            return;
        }
        Map<String, List<String>> categoryMap = categoryList.stream().collect(Collectors.groupingBy(CrawlerMetaArticle::getUniqueIndex,
                Collectors.mapping(CrawlerMetaArticle::getCategory, Collectors.toList())));
        for (Content content : contentList) {
            String md5 = articleMd5Map.get(content.getId());
            content.setCategory(categoryMap.get(md5));
        }
        log.info("setContentCategory cost:{}", System.currentTimeMillis() - start);
    }

    private List<CrawlerMetaArticle> getByUniqueIndexIn(List<String> md5List) {
        if (CollectionUtils.isEmpty(md5List)) {
            return new ArrayList<>();
        }
        return crawlerMetaArticleRepository.getByUniqueIndexIn(md5List);
    }


    public static String generateArticleUniqueMd5(String url) {
        // Extract parts from the URL
        String biz = extractParameter(url, "biz=");
        String idx = extractParameter(url, "&idx=");
        String sn = extractParameter(url, "&sn=");

        // Combine the parts and encode them to bytes
        String urlBit = String.format("%s-%s-%s", biz, idx, sn);
        return Md5Util.encoderByMd5(urlBit);
    }

    private static String extractParameter(String url, String parameter) {
        int start = url.indexOf(parameter) + parameter.length();
        if (start == -1 || start == url.length()) {
            return "";
        }
        int end = url.indexOf("&", start);
        if (end == -1) {
            return url.substring(start);
        } else {
            return url.substring(start, end);
        }
    }

    public void setTitleAvgViewCount(List<Content> contentList, String ghId) {
        long start = System.currentTimeMillis();
        // 获取账号相关性
        List<AccountCorrelation> accountCorrelationList = accountCorrelationRepository.findByGhIdAndStatus(ghId, 1);
        Map<String, Double> accountCorrelationMap = accountCorrelationList.stream().collect(
                Collectors.toMap(AccountCorrelation::getRelGhId, AccountCorrelation::getCorrelation));

        Set<String> titleMd5List = contentList.stream().map(o -> Md5Util.encoderByMd5(o.getTitle())).collect(Collectors.toSet());
//        Set<String> titleList = contentList.stream().map(Content::getTitle).collect(Collectors.toSet());
//        Set<String> crawlerTitleList = contentList.stream().map(Content::getCrawlerTitle).collect(Collectors.toSet());
//        titleList.addAll(crawlerTitleList);
        // 获取历史已发布文章
        List<Article> hisArticleList = new ArrayList<>();
        List<List<String>> titleMd5Partition = Lists.partition(new ArrayList<>(titleMd5List), 1000);
        for (List<String> titleMd5s : titleMd5Partition) {
            hisArticleList.addAll(articleRepository.getByTitleMd5InAndTypeEqualsAndStatusEquals(titleMd5s, "9", 1));
        }
        Map<String, Map<Integer, List<Article>>> map = hisArticleList.stream()
                .collect(Collectors.groupingBy(Article::getTitle, Collectors.groupingBy(Article::getItemIndex)));
        Set<String> snList = hisArticleList.stream().map(Article::getWxSn).collect(Collectors.toSet());
        List<ArticleDetailInfo> articleDetailInfoList = new ArrayList<>();
        List<List<String>> snPartition = Lists.partition(new ArrayList<>(snList), 1000);
        for (List<String> sns : snPartition) {
            articleDetailInfoList.addAll(articleDetailInfoRepository.getAllByWxSnIn(sns));
        }
        Map<String, List<ArticleDetailInfo>> articleDetailInfoMap = articleDetailInfoList.stream()
                .collect(Collectors.groupingBy(ArticleDetailInfo::getWxSn));
        // 获取历史已发布文章所属头条内容
        Set<String> ghIds = hisArticleList.stream().map(Article::getGhId).collect(Collectors.toSet());
        Set<String> appMsgIds = hisArticleList.stream().map(Article::getAppMsgId).collect(Collectors.toSet());
        List<Article> firstIndexHisArticleList = articleRepository.getByGhIdInAndAppMsgIdInAndItemIndexAndTypeEqualsAndStatusEquals(ghIds, appMsgIds, 1, "9", 1);
        Map<String, Map<String, Article>> firstIndexHisArticleMap = firstIndexHisArticleList.stream()
                .collect(Collectors.groupingBy(Article::getGhId, Collectors.toMap(Article::getAppMsgId, o -> o)));
        // 获取发布账号 位置历史均值
        List<AccountAvgInfo> accountAvgInfoList = accountAvgInfoRepository.getAllByGhIdIn(ghIds);
        Map<String, Map<String, Map<String, AccountAvgInfo>>> accountAvgInfoIndexMap = accountAvgInfoList.stream()
                .filter(o -> Objects.nonNull(o.getReadAvg()) && o.getReadAvg() > 0 && o.getFans() > 1000)
                .collect(Collectors.groupingBy(AccountAvgInfo::getGhId, Collectors.groupingBy(AccountAvgInfo::getUpdateTime,
                        Collectors.toMap(AccountAvgInfo::getPosition, o -> o))));
        Map<String, Map<String, AccountAvgInfo>> firstIndexAvgInfoMap = accountAvgInfoList.stream()
                .filter(o -> "1".equals(o.getPosition()))
                .filter(o -> Objects.nonNull(o.getReadAvg()) && o.getReadAvg() > 0 && o.getFans() > 1000)
                .collect(Collectors.groupingBy(AccountAvgInfo::getGhId, Collectors.toMap(AccountAvgInfo::getUpdateTime, o -> o)));
        for (Content content : contentList) {
            List<Article> hisArticles = new ArrayList<>();
            Map<Integer, List<Article>> indexArticleMap = map.get(content.getTitle());
            if (Objects.isNull(indexArticleMap)) {
                indexArticleMap = map.get(content.getCrawlerTitle());
            } else if (!content.getTitle().equals(content.getCrawlerTitle())) {
                Map<Integer, List<Article>> crawlerTitleIndexArticleMap = map.get(content.getCrawlerTitle());
                if (Objects.nonNull(crawlerTitleIndexArticleMap)) {
                    for (Map.Entry<Integer, List<Article>> entry : crawlerTitleIndexArticleMap.entrySet()) {
                        if (indexArticleMap.containsKey(entry.getKey())) {
                            List<Article> articleList = indexArticleMap.get(entry.getKey());
                            articleList.addAll(entry.getValue());
                            indexArticleMap.put(entry.getKey(), articleList);
                        } else {
                            indexArticleMap.put(entry.getKey(), entry.getValue());
                        }
                    }
                }
            }
            if (Objects.nonNull(indexArticleMap) && !indexArticleMap.isEmpty()) {
                for (Map.Entry<Integer, List<Article>> indexArticleEntry : indexArticleMap.entrySet()) {
                    List<Article> indexArticleList = indexArticleEntry.getValue();
                    hisArticles.addAll(indexArticleList);
                }
            }
            content.setHisPublishArticleList(new ArrayList<>());
            for (Article hisArticle : hisArticles) {
                ContentHisPublishArticle article = new ContentHisPublishArticle();
                BeanUtils.copyProperties(hisArticle, article);
                article.setViewCount(hisArticle.getShowViewCount());
                article.setWxSn(hisArticle.getWxSn());
                article.setArticleDetailInfoList(articleDetailInfoMap.get(hisArticle.getWxSn()));
                // 设置账号位置阅读均值
                int avgViewCount = 0;
                Map<String, Map<String, AccountAvgInfo>> dateAvgMap = accountAvgInfoIndexMap.get(hisArticle.getGhId());
                String hisPublishDate = DateUtils.timestampToYMDStr(article.getUpdateTime(), "yyyy-MM-dd");
                if (Objects.nonNull(dateAvgMap)) {
                    List<String> avgMapDateList = new ArrayList<>(dateAvgMap.keySet());
                    hisPublishDate = DateUtils.findNearestDate(avgMapDateList, hisPublishDate, "yyyy-MM-dd");
                    Map<String, AccountAvgInfo> indexMap = dateAvgMap.get(hisPublishDate);
                    if (Objects.nonNull(indexMap) && indexMap.containsKey(hisArticle.getItemIndex().toString())) {
                        article.setInnerAccount(true);
                        avgViewCount = Optional.ofNullable(indexMap.get(hisArticle.getItemIndex().toString()).getReadAvg())
                                .orElse(0.0).intValue();
                    } else {
                        FeishuMessageSender.sendWebHookMessage("07026a9f-43f5-448b-ba40-a8d71bd6e634", "历史表现阅读均值获取失败\n"
                                + "ghId: " + hisArticle.getGhId() + "\n"
                                + "账号名称: " + hisArticle.getAccountName() + "\n"
                                + "日期: " + hisPublishDate + "\n"
                                + "位置: " + hisArticle.getItemIndex());
                    }
                }
                article.setAvgViewCount(avgViewCount);
                if (Objects.nonNull(article.getAvgViewCount()) && article.getAvgViewCount() > 0
                        && Objects.nonNull(article.getViewCount())) {
                    article.setViewCountRate((article.getViewCount() * 1.0) / article.getAvgViewCount());
                }
                // 设置头条阅读均值
                AccountAvgInfo firstIndexAvgInfo = null;
                Map<String, AccountAvgInfo> firstIndexDateAvgInfo = firstIndexAvgInfoMap.get(hisArticle.getGhId());
                if (Objects.nonNull(firstIndexDateAvgInfo)) {
                    firstIndexAvgInfo = firstIndexDateAvgInfo.get(hisPublishDate);
                    if (Objects.nonNull(firstIndexAvgInfo)) {
                        article.setFans(firstIndexAvgInfo.getFans());
                    }
                }
                Map<String, Article> firstIndexArticle = firstIndexHisArticleMap.get(hisArticle.getGhId());
                if (Objects.nonNull(firstIndexArticle) && firstIndexArticle.containsKey(hisArticle.getAppMsgId())) {
                    Article firstArticle = firstIndexArticle.get(hisArticle.getAppMsgId());
                    article.setFirstViewCount(firstArticle.getShowViewCount());
                    if (Objects.nonNull(firstIndexAvgInfo) && Objects.nonNull(firstIndexAvgInfo.getReadAvg())
                            && firstIndexAvgInfo.getReadAvg() > 0 && Objects.nonNull(firstArticle.getShowViewCount())) {
                        article.setFirstViewCountRate((firstArticle.getShowViewCount() * 1.0) / firstIndexAvgInfo.getReadAvg());
                    }
                }
                article.setCorrelation(Optional.ofNullable(accountCorrelationMap.get(article.getGhId())).orElse(0.0));
                content.getHisPublishArticleList().add(article);
            }
            // 设置头条阅读均值
            setT0Data(content);
        }
        log.info("setTitleAvgViewCount cost:{}", System.currentTimeMillis() - start);
    }

    private void setT0Data(Content content) {
        if (CollectionUtils.isEmpty(content.getHisPublishArticleList())) {
            return;
        }
        int firstLevelSize = 0;
        int fissionSum = 0;
        double fissionWeightSum = 0;
        int fansSum = 0;
        int avgReadCountSum = 0;
        Double t0FissionByFansSum = 0.0;
        Double t0FissionByReadAvgSum = 0.0;
        Double t0FissionByReadAvgCorrelationSum = 0.0;
        for (ContentHisPublishArticle article : content.getHisPublishArticleList()) {
            if (article.getItemIndex() != 1 || !article.isInnerAccount()) {
                continue;
            }
            if (CollectionUtils.isEmpty(article.getArticleDetailInfoList())) {
                // 仅判断7.12以后发布文章
                if (article.getUpdateTime() > 1720713600 && contentHisFeishuEnable) {
                    FeishuMessageSender.sendWebHookMessage("07026a9f-43f5-448b-ba40-a8d71bd6e634", "历史表现裂变特征获取失败\n"
                            + "ghId: " + article.getGhId() + "\n"
                            + "账号名称: " + article.getAccountName() + "\n"
                            + "位置: " + article.getItemIndex() + "\n"
                            + "标题: " + article.getTitle() + "\n"
                            + "发布时间: " + DateUtils.timestampToYMDStr(article.getUpdateTime(), "yyyyMMdd") + "\n"
                            + "wxsn: " + article.getWxSn());
                }
                continue;
            }
            int sumFission0 = 0;
            Date minDate = article.getArticleDetailInfoList().stream().map(ArticleDetailInfo::getRecallDt).min(Date::compareTo).orElse(new Date());
            for (ArticleDetailInfo articleDetailInfo : article.getArticleDetailInfoList()) {
                if (articleDetailInfo.getRecallDt().equals(minDate) && Objects.nonNull(articleDetailInfo.getFission0())) {
                    sumFission0 += articleDetailInfo.getFission0();
                }
            }
            if (sumFission0 == 0) {
                continue;
            }
            Double correlation = (article.getCorrelation() + 1) / 2;
            article.setT0FissionSum(sumFission0);
            if (article.getFans() > 0) {
                article.setT0FissionByFans(sumFission0 * 1.0 / article.getFans());
                fansSum += (int) article.getFans();
                t0FissionByFansSum += article.getT0FissionByFans();
            }
            if (Objects.nonNull(article.getAvgViewCount()) && article.getAvgViewCount() > 0) {
                article.setT0FissionByReadAvg(sumFission0 * 1.0 / (article.getAvgViewCount() + 500));
                avgReadCountSum += article.getAvgViewCount();
                t0FissionByReadAvgSum += article.getT0FissionByReadAvg();
                t0FissionByReadAvgCorrelationSum += article.getT0FissionByReadAvg() * correlation;
            }
            fissionSum += sumFission0;
            int hour = DateUtils.getHourByTimestamp(article.getUpdateTime());
            if (hour >= 12) {
                fissionWeightSum += sumFission0 / morningNoonFissionRate;
            } else {
                fissionWeightSum += sumFission0;
            }
            firstLevelSize++;
        }
        if (firstLevelSize > 0) {
            content.setT0FissionByFansMean(t0FissionByFansSum / firstLevelSize);
            content.setT0FissionByReadAvgMean(t0FissionByReadAvgSum / firstLevelSize);
            content.setT0FissionByReadAvgCorrelationMean(t0FissionByReadAvgCorrelationSum / firstLevelSize);
            if (fansSum > 0) {
                content.setT0FissionByFansSumAvg(fissionSum * 1.0 / fansSum);
            }
            if (avgReadCountSum > 0) {
                content.setT0FissionByReadAvgSumAvg(fissionSum * 1.0 / avgReadCountSum);
            }
            if (avgReadCountSum > 0) {
                content.setT0FissionDeWeightByReadAvgSumAvg(fissionWeightSum / avgReadCountSum);
            }
        }
    }

    public static void main(String[] args) {
        String url = "http://mp.weixin.qq.com/s?__biz=Mzg2ODk4MTg3OQ==&mid=2247488306&idx=1&sn=93ebadc5bc7161a0dee48355013d3bc4&chksm=cfb6c1cb2bcdd80dd16d5d604d741a0019ae791125265a042d26100ba21ddb9e5c643ecc2264&scene=126&sessionid=1679649075#rd";
        String md5 = generateArticleUniqueMd5(url);
        System.out.println("Generated md5: " + md5);
    }

}
